package logic;

import java.util.ArrayList;
import java.util.Hashtable;

public class SequenceWithRibosomes {
	private int nextRibosome;
	private int elapsedTime = 0;
	private final int FI_PRECISION = 1000;
	private final int RIBOSOME_RADIUS = 5;
	// private boolean simulationFinished = false;

	private String sequence;
	private String proteinSequence;
	private boolean homoSapiens;
	private ArrayList<String> codons = new ArrayList<String>();
	private ArrayList<Integer> ribosomePositions = new ArrayList<Integer>();
	private ArrayList<Integer> ribosomeCountdowns = new ArrayList<Integer>();
	// private int nextRibosomeCountDown;
	private ArrayList<Integer> clashes = new ArrayList<Integer>();
	private Hashtable<String, Integer> delays = new Hashtable<String, Integer>();
	private Hashtable<Integer, String> originalCodons = new Hashtable<Integer, String>();

	public void setDataFromParams(String seq, String next_delay,
			String delayCSV, boolean homoSapiensParam) {
		sequence = seq;
		nextRibosome = (int) (FI_PRECISION * Double.parseDouble(next_delay));
		codons = new ArrayList<String>();
		String codon;
		for (int i = 0; i < sequence.length(); i += 3) {
			codon = sequence.substring(i, i + 3);
			codons.add(codon);
		}
		StringBuilder proteinSequenceBuilder = new StringBuilder();
		for (String c : codons) {
			proteinSequenceBuilder.append(Translator.getInstance()
					.translateCodon(c));
		}
		proteinSequence = proteinSequenceBuilder.toString();
		// delays for specific codons
		for (String d : delayCSV.split(";")) {
			String[] kv = d.split(",");
			delays.put(kv[0], Integer.parseInt(kv[1]));
		}

		homoSapiens = homoSapiensParam;
		if (homoSapiens) {
			Translator.getInstance().makeHuman();
		}
	}

	private int getDelay(int codonIndex) {
		try {
			if (codons.get(codonIndex) != "UGA") {
				return delays.get(codons.get(codonIndex));
			} else { // it might be selenocysteine
				return (codonIndex != codons.size()) ? delays.get("UGA") : 0;
			}
		} catch (IndexOutOfBoundsException e) { // stop codon
			return 0;
		}
	}

	public void simulationStart() {
		// set first ribosome on position 0
		elapsedTime = 0;
		ribosomePositions = new ArrayList<Integer>();
		ribosomePositions.add(0);
		// set time it will wait at this position
		ribosomeCountdowns = new ArrayList<Integer>();
		ribosomeCountdowns.add(getDelay(0));
		// set time it will wait for next ribosome start
		// nextRibosomeCountDown = nextRibosome;
		// reset clashes table
		clashes = new ArrayList<Integer>();
		// simulationFinished = false;
	}

	public boolean simulationStep() {
		elapsedTime++;
		// return true if something changed
		boolean ans = false;
		// one step means one millisecond
		// check if we should move ribosomes
		for (int i = 0; i < ribosomeCountdowns.size(); i++) {
			ribosomeCountdowns.set(i, ribosomeCountdowns.get(i) - 1);
			if (ribosomeCountdowns.get(i) == 0) {
				ans = true;
				// move ribosome forward
				ribosomePositions.set(i, ribosomePositions.get(i) + 1);
				ribosomeCountdowns.set(i, getDelay(ribosomePositions.get(i)));
			}
		}
		// check if it's on the end
		for (int i = 0; i < ribosomeCountdowns.size(); i++) {
			if (ribosomePositions.get(i) >= codons.size() - 1) {
				ribosomePositions.remove(i);
				ribosomeCountdowns.remove(i);
				ans = true;
				// simulationFinished = true;
			}
		}
		// check if we should add a new ribosome
		if (elapsedTime % (nextRibosome) == 0) {
			// do it
			ans = true;
			ribosomePositions.add(0);
			ribosomeCountdowns.add(getDelay(0));
		}
		// check for clashes (straightforward, primitive way)
		for (int i = 0; i < ribosomePositions.size(); i++) {
			for (int j = 0; j < ribosomePositions.size(); j++) {
				if (Math.abs(ribosomePositions.get(i)
						- ribosomePositions.get(j)) < (RIBOSOME_RADIUS * 2 + 1)
						&& i != j) {
					clashes.add(ribosomePositions.get(i));
					ans = true;
				}
			}
		}
		return ans;
	}

	public boolean simulationStepBackwards() {
		if (elapsedTime == 0)
			return false;
		elapsedTime--;
		// return true if something changed
		boolean ans = false;
		// one step means one millisecond
		// check if we should move ribosomes
		for (int i = 0; i < ribosomeCountdowns.size(); i++) {
			ribosomeCountdowns.set(i, ribosomeCountdowns.get(i) + 1);
			if (ribosomeCountdowns.get(i) == getDelay(ribosomePositions.get(i))) {
				ans = true;
				// move ribosome backward
				ribosomePositions.set(i, ribosomePositions.get(i) - 1);
				ribosomeCountdowns.set(i, 0);
			}
		}
		// check if it's on the beginning
		for (int i = 0; i < ribosomeCountdowns.size(); i++) {
			if (ribosomePositions.get(i) <= 0) {
				ans = true;
				// simulationFinished = true;
			}
		}
		// check for clashes (straightforward, primitive way)
		clashes = new ArrayList<Integer>();
		for (int i = 0; i < ribosomePositions.size(); i++) {
			for (int j = 0; j < ribosomePositions.size(); j++) {
				if (Math.abs(ribosomePositions.get(i)
						- ribosomePositions.get(j)) < (RIBOSOME_RADIUS * 2 + 1)
						&& i != j) {
					clashes.add(ribosomePositions.get(i));
					ans = true;
				}
			}
		}
		return ans;
	}

	public boolean multipleSimulationSteps(int number) {
		boolean ans = false;
		if (number > 0) {
			for (int i = 0; i < number; i++) {
				ans = simulationStep() || ans;
			}
		} else {
			for (int i = 0; i < -number; i++) {
				ans = simulationStepBackwards() || ans;
			}
		}
		return ans;
	}

	public void simulationClean() {
		elapsedTime = 0;
		ribosomePositions = new ArrayList<Integer>();
		ribosomeCountdowns = new ArrayList<Integer>();
		clashes = new ArrayList<Integer>();
	}

	public char charAt(int index) {
		return sequence.charAt(index);
	}

	public boolean[] getRibosomes() {
		boolean[] ans = new boolean[codons.size()];
		for (int i = 0; i < ans.length; i++)
			ans[i] = false;
		for (int i : ribosomePositions) {
			try {
				ans[i] = true;
			} catch (ArrayIndexOutOfBoundsException e) {

			}
		}
		return ans;
	}

	public boolean[] getClashes() {
		boolean[] ans = new boolean[codons.size()];
		for (int i = 0; i < ans.length; i++)
			ans[i] = false;
		for (int i : clashes) {
			try {
				ans[i] = true;
			} catch (ArrayIndexOutOfBoundsException e) {

			}
		}
		return ans;
	}

	public ArrayList<String> getCodons() {
		return codons;
	}

	public void mutate(int index, String newCodon) {
		if (originalCodons.containsKey(index)) { // it's already modified
			if (originalCodons.get(index).replace("T", "U").equals(newCodon)) { // back
																				// to
				// original
				originalCodons.remove(index);
			}
		} else { // first mutation
			originalCodons.put(index, codons.get(index).replace("T", "U"));
		}
		sequence = sequence.substring(0, index * 3) + newCodon
				+ sequence.substring((index + 1) * 3);
		codons.set(index, newCodon);
		proteinSequence = proteinSequence.substring(0, index)
				+ Translator.getInstance().translateCodon(newCodon)
				+ proteinSequence.substring(index + 1);
	}

	public int length() {
		try {
			return sequence.length();
		} catch (NullPointerException e) {
			return 0;
		}
	}

	public String toString() {
		if (!sequence.isEmpty())
			return sequence;
		else
			return "";
	}

	public String getProteinSequence() {
		return proteinSequence;
	}

	public boolean anyClashes() {
		return (clashes.size() > 0);
	}

	public Hashtable<Integer, String> getOriginalCodons() {
		return originalCodons;
	}

	public int getElapsedTime() {
		return elapsedTime;
	}

	public void setElapsedTime(int elapsedTime) {
		this.elapsedTime = elapsedTime;
	}

	/*
	 * public boolean isSimulationFinished() { return simulationFinished; }
	 */

	public double getDelay(String codon) {
		return delays.get(codon.replace("T", "U"));
	}

	public String getSequence() {
		return sequence;
	}
}
